
/*
   Copyright (c) 2002 Compaq Computer Corporation
   
   SOFTWARE RELEASE
   
   Permission is hereby granted, free of charge, to any person obtaining
   a copy of this software and associated documentation files (the
   "Software"), to deal in the Software without restriction, including
   without limitation the rights to use, copy, modify, merge, publish,
   distribute, sublicense, and/or sell copies of the Software, and to
   permit persons to whom the Software is furnished to do so, subject to
   the following conditions:
   
   - Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
   
   - Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
   
   - Neither the names of Compaq Research, Compaq Computer Corporation
     nor the names of its contributors may be used to endorse or promote
     products derived from this Software without specific prior written
     permission.
   
   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
   IN NO EVENT SHALL COMPAQ COMPUTER CORPORATION BE LIABLE FOR ANY CLAIM,
   DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
   OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
   THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

package AccordionTreeDrawer;

import java.awt.*;
import java.io.IOException;
import java.util.*;
import net.java.games.jogl.GL;

import AccordionDrawer.*;

/**
 * A class representing a node of a (phylognenetic) tree.
 * The tree that this node belongs to is of type Tree.
 * Nodes have fields that store a pre- and post-ordering. 
 *
 * A TreeNode has a list of children, a unique key, a leftmostleaf 
 * and a rightmost leaf
 *
 * @author  Tamara Munzner, Li Zhang, Yunhong Zhou
 * @version 
 * @see     Tree
 * @see     GridCell
 */

public class TreeNode extends Object implements CellGeom, Comparable {

    ArrayList children; // eventually turn this into an array (need to change parser)
//    TreeEdge edges[] = new TreeEdge[2];
//    LabelBox lb;

    /** key is unique for nodes in one tree */
    public int key; 

    /** The GridCell that this node is attached to  */ 
    public GridCell cell; 

    /** 
     * The index of the smallest descendant node (Node in subtree rooted
     * at this node).
     */
    private int min; 

    /** The index of the largest descendant node */ 
    private int max; 
    
    private int fontSize;
    Double bcnScore;
//    /**
//     * The offset of the point with respect to the cell. We only have
//     * this for the row offset as we assume that the vertical edges
//     * are all aligned.  When computing the Y coordinate of a node, we
//     * add nodeOffsetR to the pointOffset[1], a fixed parameter set by
//     * AccordionDrawer.
//     */
	int computedFrame; // store frame midYPosition was last calculated (needed to place parents)
    double midYPosition;


    /** @return The index of the smallest descendant node */ 
    public int getMin() { return min; }

    /** @return The index of the smallest descendant node */ 
    public int getMax() { return max; }

    public int getKey() { return key;}
    public String getName() { return name;}

    public boolean pick(int x, int y) { return false; }

    /**
     * Draws this node inside cell, the GridCell that it is attached
     * to, if cell is drawn large enough Prints the label of the node
     * if the GridCell is drawn large enough
     * @author   Tamara Munzner, Serdar Tasiran, Li Zhang, Yunhong Zhou
     * 
     * @param    col The color to draw this node in
     * @param    plane The plane to draw this node in
     *
     * @see      AccordionDrawer.GridCell
     */
    
    public void drawInCell(Color col, double plane) {
	for (int xy = 0; xy < 2; xy++) {
if (getEdge(xy)) drawInCell(col, plane, xy==0);	}
    }
    
    public void drawInCell(ArrayList col, double plane)
    {
//        System.out.println("drawInCell top " + this);
    	if (cell.getDrawnFrame() >= cell.drawer.getFrameNum())
    		return; // already drawn in this frame
    	cell.setDrawnFrame(cell.drawer.getFrameNum());
    	if (isLeaf())
    		((AccordionTreeDrawer)cell.drawer).leafDrawCount++;
    	else
			((AccordionTreeDrawer)cell.drawer).internalDrawCount++;
    	Color c = null;
    	if (col != null && col.size() > 0)
    	{
    		c = (Color)col.get(0);
    		plane = cell.drawer.hiliteplane - .004*c.getAlpha();
    	}
    	drawInCell(c, plane);
    }

    // cell max[X] boundary set to min[X] boundary over all children
    public void setCell(GridCell c) { cell = c;}
    public GridCell getCell() { return cell;}

    public boolean getEdge(int xy) { if (xy==0) return !isRoot(); else return !isLeaf();}

    /** implement Comparable interface - sort on key field */
    public int compareTo(Object o) {
	if (key == ((TreeNode)o).key) return 0;
	else if (key < ((TreeNode)o).key) return -1;
	else return 1;
    }

    public TreeNode parent;

    /** 
     * node name with default "". 
     * Most internal nodes have no name and all leaf nodes have a name 
     */  
    String name = ""; // the long form in fully qualified names
    public String label = ""; // always short form

    int height;

    /** weight is actually edge length for the edge immediately above the node */
    public float weight = 0.0f;


	/**
	 * Draw label of this TreeEdge at maximum size (intended for
	 * mouseover highlighting).
	 * 
	 * @author   Tamara Munzner
	 */
	
	public void drawLabelBig(int x, int y, int fontheight, boolean horiz) {
		AccordionTreeDrawer drawer = (AccordionTreeDrawer) getCell().drawer;
		if (!horiz || drawer.nogeoms || label.length() < 1)
			return;
		drawLabelBox(drawer.flashLabelBox, fontheight, true);
	}

    public TreeNode leftmostLeaf, rightmostLeaf;
    public int numberLeaves;
    public TreeNode preorderNext = null;

    public TreeNode posorderNext = null;

// this can be extracted from split cell position
//    /** leaf index */
//    public int lindex; 

// change next pointers to list of leaves with leaf indices corresponding to
// position in split cells
//    /** next leaf added by Li */
//    public TreeNode next=null;


    /**
     * Children list initial capacity 2 as in most case binary.
     */
    // used in 2 places:
    // - create the root when creating the tree
    // - the parser uses this to create nodes attached to the root
    public TreeNode() {

	children = new ArrayList(2); 
	max = -1;
	min = Integer.MAX_VALUE;
	bcnScore = new Double(0.0);
	//	nodeOffsetR = 0;
    }

	public void drawInCell(Color col, double plane, boolean horiz) {
			//    	System.out.println("Drawing: " + node.getCell());
			int X = horiz ? 0 : 1;
			int Y = horiz ? 1 : 0;
			AccordionTreeDrawer atd = (AccordionTreeDrawer) cell.drawer;
//			System.out.println("drawInCell " + this);
			boolean isBase = false;
			if (col == null)
			{
				col = atd.objectColor;
				isBase = true;
			}
			
			if (atd.nogeoms)
				return;
			AccordionTreeDrawer.countDrawnFrame++;
			GL gl = (GL) atd.getGraphicsStuff(); // this is not a gl call, just a reference
			Color drawColor = setGLColor(atd, col);
	
			// draw the line
			int thick = atd.getLineThickness();
			if (col != atd.getObjectColor() && !atd.isDoingFlash())
				thick += 2;
			float[] posStart = new float[2], posEnd = new float[2];
			setPositions(posStart, posEnd, horiz);//, atd.drawDirection);
			//	if (horiz)
			//		System.out.println(node + " Set positions: Y:" + posEnd[1]);
			if (atd.takeSnapshot)
			{
				if ((atd.basePass && isBase) || (atd.groupPass && !isBase))
				try
				{
					int cellMinPix[] = {atd.w2s(posStart[X], X), atd.w2s(posStart[Y], Y)};
					int cellMaxPix[] = {atd.w2s(posEnd[X], X), atd.w2s(posEnd[Y], Y)};
					atd.snapShotWriter.write(thick + " setlinewidth newpath " +
							cellMinPix[X] + " " + cellMinPix[Y] + " moveto " +
							cellMaxPix[X] + " " + cellMaxPix[Y] + " lineto " +
							"closepath " +
							"gsave " + 
							drawColor.getRed() / 255f + " " +
							drawColor.getGreen() / 255f + " " +
							drawColor.getBlue() / 255f + " setrgbcolor " +
							"stroke grestore\n");
				}
				catch (IOException ioe)
				{
					System.out.println("Error: IOException while trying to write cell to file: " + atd.snapShotWriter.toString());
				}
			
			}
			else
			{
				atd.setColorGL(drawColor);
				gl.glLineWidth(thick);
				gl.glBegin(GL.GL_LINES);
//				System.out.println("drawing " + " " +
//				        posStart[0] + " " + posStart[1] + " " +
//				        posEnd[0] + " " + posEnd[1]);
				gl.glVertex3d(posStart[0], posStart[1], plane);
				gl.glVertex3d(posEnd[0], posEnd[1], plane);
				gl.glEnd();
			}
	
			if (horiz)
			{ // draw the dot at the end of the horizontal section
				if (atd.takeSnapshot)
				{
					if ((atd.basePass && isBase) || (atd.groupPass && !isBase))
					try
					{
						int cellMaxPix[] = {atd.w2s(posEnd[X], X), atd.w2s(posEnd[Y], Y)};
						atd.snapShotWriter.write("newpath " +
								(cellMaxPix[X]-thick) + " " + (cellMaxPix[Y]-thick) + " moveto " +
								(cellMaxPix[X]-thick) + " " + (cellMaxPix[Y]+thick) + " lineto " +
								(cellMaxPix[X]+thick) + " " + (cellMaxPix[Y]+thick) + " lineto " +
								(cellMaxPix[X]+thick) + " " + (cellMaxPix[Y]-thick) + " lineto " +
								"closepath " +
								"gsave " + 
								drawColor.getRed() / 255f + " " +
								drawColor.getGreen() / 255f + " " +
								drawColor.getBlue() / 255f + " setrgbcolor " +
								"eofill grestore\n");
					}
					catch (IOException ioe)
					{
						System.out.println("Error: IOException while trying to write cell to file: " + atd.snapShotWriter.toString());
					}
				
				}
				else
				{
					gl.glPointSize((float) (thick + 2.0f));
					gl.glBegin(GL.GL_POINTS);
					gl.glVertex3d(posEnd[0], posEnd[1], plane);
					gl.glEnd();
				}
			}
	
			if (horiz && atd.drawlabels && label.length() > 0 && !atd.basePass && !atd.groupPass ) {
				// first, check smallest font, to see if anything at all
				// can fit. if it does, then do binary search to find
				// largest possible box that fits. after we know which box
				// size to use, draw it.
	
				// this algorithm means drawing order will have a radical
				// effect on how labels look!		
					
					// we should set min to one less than the real bound. new
					// real bound is minFontSize+1 (because we already tested
					// minFontSize), so leave min set to current value of
					// (minFontSize+1)-1 = minFontSize
					
					int min = ((AccordionTreeDrawer)atd).minFontHeight;
					int max = ((AccordionTreeDrawer)atd).maxFontHeight;
					int mid = max;
					int fitsheight = min;
					LabelBox fits = makeLabelBox(min, -1, -1, posStart, posEnd);
	//				System.out.println("min font size: " + min);
					if (!intersectLabelBox(fits))
					{
						if (max != min)
						{
							LabelBox maxBox = makeLabelBox(max, -1, -1, posStart, posEnd);
							if (!intersectLabelBox(maxBox))
							{
								fits = maxBox;
								fitsheight = max;
							}
							else while (min+1 < max)
							{
								mid = (int) ((min + max) / 2.0);
								LabelBox lb = makeLabelBox(mid, -1, -1, posStart, posEnd);
								if (intersectLabelBox(lb))
									max = mid; // too big
								else
								{
									fits = lb; // bigger size
									fitsheight = mid;
									min = mid;
								}
							}
						}
	//					if (atd.takeSnapshot)
	//						System.out.println(node.name + " using font size: " + fitsheight);
						drawLabelBox(fits, fitsheight, false);
	//					else
	//						drawLabelBox(fits, fitsheight, false);
					}
			}
		}

	/**
	 * clean the node itself and the tree edge attached with it.
	 * @see TreeEdge.close
	 * @see Tree.close
	 *
	 */   
	   public void close(){
		   children.clear();
		   children=null;
		   name=null;
		   cell = null;
			parent=null;	
			label = null;
			leftmostLeaf=null;
			rightmostLeaf=null;	
			preorderNext = null;
			posorderNext = null;		
			bcnScore=null;
//		   this = null;
		  }
   
		  protected void finalize() throws Throwable {
		 
					   try {
						   close();
					   }
					   finally {
						   super.finalize();     
//						   System.out.println("finally clean treeNodes");
					   }
		   }

    /**
     * Allocate a new TreeNode object with node name and weight.   
     * @param s The name of the node given by a string
     * @param w The weight of the edge immediately above the node
     */
    // used only in copying trees
    public TreeNode(String s, float w){
	children = new ArrayList(2);
	name = new String(s);
	weight = w;
	max = -1;
	min = Integer.MAX_VALUE;
	// nodeOffsetR = 0;
    }

	// not used
    public TreeNode(String s){
	children = new ArrayList(2);
	name = new String(s);
	max = -1;
	min = Integer.MAX_VALUE;
	// nodeOffsetR = 0;
    }

    public void setName(String s){
	name = new String(s);
    }

    public int numberChildren() {
	return children.size();
    }

    public TreeNode getChild(int i){
	return (TreeNode) children.get(i);
    }

    public boolean isLeaf() {
	return children.isEmpty();
    }

    public boolean isRoot() {
	return (null == parent);
    }


    /** @return the total number of leaves decedant to this node */
    public int getLeafCount() {	
	if(this == null)
	    return 0;
	else if(isLeaf())
	    return 1;
	else{
	    int sum = 0;
	    for(int i=0; i<numberChildren(); i++)
		sum += getChild(i).getLeafCount();
	    return sum;
	}
	
    }

    public boolean equals(TreeNode n){
	return (name.equals(n.name));
    }

    public void addChild(TreeNode n) {
	children.add(n);
	n.parent = this;
    }

    public TreeNode parent() {
	return parent;
    }

    public void setWeight(double w){
	weight = (float) w;
    }

    public float getWeight() { return weight; }
    
    public TreeNode firstChild() {
	return (TreeNode) children.get(0);
    }

    public TreeNode lastChild() {
	return (TreeNode) children.get(children.size()-1);
    }

    /** 
     * @return <code>true</code> if this is an ancestor of <code>that</code>
     * @param that Another TreeNode object 
     */
    public boolean isAncestorOf(TreeNode that) {
	if(leftmostLeaf.getLindex() > that.leftmostLeaf.getLindex() ||
	  	rightmostLeaf.getLindex() < that.rightmostLeaf.getLindex()) return false;
	return true;
	
    }

    /**
     * Compute the lowest common ancestor between this node and <code>that</that>.
     * The two nodes must belong to the same tree.
     * @author   Li Zhang
     * 
     * @param    that A TreeNode in the Tree that this TreeNode belongs to
     * @return   the lowest common ancestor between this node and "that"
     */
    
    public TreeNode lca(TreeNode that) {
	if(that.isAncestorOf(this)) return that;
	TreeNode current = this;
	while(current != null) {
	    if(current.isAncestorOf(that)) return current;
	    current = current.parent;
	}
	return null;
    }


    /**
     * Compute the lowest common ancestor between this leaf and "that"
     * The two nodes must belong to the same tree and must be leaves
     * @author   Li Zhang
     * 
     * @param    that A TreeNode in the Tree that this TreeNode belongs to
     * @return   the lowest common ancestor between this leaf and <code>that</code>, null if one of the nodes is not a leaf 
     */
    public TreeNode leafLca(TreeNode that) {


	if(!isLeaf()) return null;
	if(!that.isLeaf()) return null;
	TreeNode current = this;
	int a = that.getLindex();
	if(getLindex() > that.getLindex()) {
	    current = that;
	    a = getLindex();
	}
	for(;current!=null;current = current.parent) {
	    if(current.rightmostLeaf.getLindex()>=a)
		return current;
	}
	return null;
    }

    public void print() {
	if(name != null)
	    System.out.print("node name: " + name + "\t");
	else 
	    System.out.print("node name null,\t");
	System.out.println("key: " + key);
    }

    public void printSubtree(){

	print();
	
	for(int i=0; i< children.size(); i++)
	    getChild(i).printSubtree();
    }
       
    public void setSubtreeExtremeLeaves() {
	if(isLeaf()) {
	    leftmostLeaf = this;
	    rightmostLeaf = this;
	    return;
	}

	int i;
	for(i=0; i<children.size();i++) {
	    getChild(i).setSubtreeExtremeLeaves();
	}
	leftmostLeaf = firstChild().leftmostLeaf;
	rightmostLeaf = lastChild().rightmostLeaf;
    }

    
    void setSubtreeMinMax() {
	if(isLeaf()) {
	    min = key;
	    max = key;
	    return;
	}
	for(int i=0; i<children.size();i++) {
	    getChild(i).setSubtreeMinMax();
	    int childMin = getChild(i).min;
	    min = (childMin < min) ? childMin : min;
	    int childMax = getChild(i).max;
	    max = (childMax > max) ? childMax : max;
	}
    }

    /** set numberLeaves field for each subnode */
    int setSubtreeNumberLeaves()
    {
    	numberLeaves = 0;
    	if(isLeaf())
    		numberLeaves = 1; 
    	else
    		for(int i=0; i<children.size(); i++)
    			numberLeaves += getChild(i).setSubtreeNumberLeaves();
    	return numberLeaves;
    }
    
    public String toString()
    {
//    	String edge[] = {edges[0]!=null?edges[0].toString():"X", edges[1]!=null?edges[1].toString():"Y"};
    	return name + "(" + key + " @ " + height + ")";
    }
    
	public void setFontSize(int fs){ fontSize=fs;}
	   public int getFontSize(){ return fontSize;}
    
	   public boolean isInteriorNode() {
		   return children.size()>0?true:false;
	   }

    public void setBcnScore(float n){
    	bcnScore =new Double(n);
    }
    
    public Double getBcnScore(){
    	return bcnScore;
    }
    
    // where the horizontal line is drawn
    // this CAN'T be recursive, that would be too expensive
    public double getMidY()
    {
    	if (computedFrame >= cell.drawer.getFrameNum())
    		return midYPosition;
    	if (isLeaf())
    	{
    		midYPosition = getSize(AccordionDrawer.Y) * 0.5 + cell.getMin(AccordionDrawer.Y);
    	}
		else if (children.size() == 1)
		{
			midYPosition = ((TreeNode)children.get(0)).getMidY();
		}
		else
		{
    		// halfway between the offsets of end children:
    		TreeNode child0 = (TreeNode)children.get(0);
    		TreeNode childN = (TreeNode)children.get(children.size() - 1);
			midYPosition = (child0.cell.getMax(AccordionDrawer.Y) + childN.cell.getMin(AccordionDrawer.Y))/2;
		}
		computedFrame = cell.drawer.getFrameNum();
		return midYPosition;
    }
    
    // where the vertical line stops
    public double getMinY()
    {
    	if (isLeaf()) return -1.0; // no vertical line for leaf
    	return ((TreeNode)children.get(0)).getMidY();
    }

	// where the vertical line stops
	public double getMaxY()
	{
		if (isLeaf()) return -1.0; // no vertical line for leaf
		return ((TreeNode)children.get(children.size()-1)).getMidY();
	}
	
	public double getSize(int xy)
	{
		return cell.getMax(xy) - cell.getMin(xy);
	}
    
	// called on the root node only
    public TreeNode pickNode(double x, double y, double xFuzz, double yFuzz)
    {
    	if (isNodePicked(x, y, xFuzz, yFuzz))
    		return this; // picked this cell
    	if (!isLeaf() && // root == leaf
    			x > cell.getMax(AccordionDrawer.X) && // x too shallow, go deeper
    			xyInRange(y, 0.0, AccordionDrawer.Y)) // y is in range, go deeper, no fuzz
    		return pickDescend(x, y, xFuzz, yFuzz);
    	return null;
    }
    
    public boolean isNodePicked(double x, double y, double xFuzz, double yFuzz)
    {
		double min[] = {cell.getMin(AccordionDrawer.X), cell.getMin(AccordionDrawer.Y)};
		double max[] = {cell.getMax(AccordionDrawer.X), cell.getMax(AccordionDrawer.Y)};
    	double midY = getMidY(); // where to draw the horizontal line
    	return (x > max[AccordionDrawer.X] - xFuzz && x < max[AccordionDrawer.X] + xFuzz &&
    			y > min[AccordionDrawer.Y] - yFuzz && y < max[AccordionDrawer.Y] + yFuzz) ||
	    		(x > min[AccordionDrawer.X] - xFuzz && x < max[AccordionDrawer.X] + xFuzz &&
	    			y > midY - yFuzz && y < midY + yFuzz);
    }
    
    private boolean xyInRange(double value, double fuzz, int xy)
    {
    	return cell.getMin(xy) - fuzz <= value && cell.getMax(xy) + fuzz >= value;
    }

    // removed recursion
    // needs to be stack based for picking to left/right of large subtrees
    public TreeNode pickDescend(double x, double y, double xFuzz, double yFuzz)
    {
    	Stack pickingStack = new Stack();
    	pickingStack.push(this);
    	while (pickingStack.size() > 0)
    	{
    		TreeNode currRoot = (TreeNode)pickingStack.pop(); // next root to check
    		if (currRoot.isNodePicked(x, y, xFuzz, yFuzz))
    			return currRoot;
    		if (currRoot.isLeaf() || // unpicked leaf
    				!currRoot.xyInRange(y, yFuzz, AccordionDrawer.Y) || // bad Y value
    				currRoot.cell.getMin(AccordionDrawer.X) > x) // bad X value
    		{
    			continue;
    		}
    		// some child must have y in range
    		int minChild = 0;
    		int maxChild = currRoot.children.size() - 1;
    		int midChild = (minChild + maxChild + 1) / 2;
    		TreeNode currChild = (TreeNode)(currRoot.children.get(midChild));
    		while (minChild != maxChild && // converged to currChild, currChild is descended or picked
    				!currChild.xyInRange(y, 0.0, AccordionDrawer.Y)) // binary search in Y, no fuzz
    		{ // find appropriate child
    			if (currChild.cell.getMin(AccordionDrawer.Y) > y) // curr child is too big, max = mid
    				maxChild = midChild;
    			else
    				minChild = midChild;
    			if (minChild+1==maxChild && midChild==minChild)
    				midChild = maxChild;
    			else
    				midChild = (minChild + maxChild) / 2;
    			currChild = (TreeNode)(currRoot.children.get(midChild));
    		}
    		if (currChild.isNodePicked(x, y, xFuzz, yFuzz))
    		{
    			return currChild; // found something
    		}
    		if (midChild > 0)
    			pickingStack.push(currRoot.children.get(midChild-1));
    		if (midChild < currRoot.children.size()-1)
    			pickingStack.push(currRoot.children.get(midChild+1));
    		pickingStack.push(currChild); // the next root
    	}
    	return null;
    }

	// assuming this is a leaf, return the leaf index (the split position in Y direction, actually)
	public int getLindex()
	{
		if (cell == null)
			System.out.println("Error when finding Lindex for " + this);
		return cell.getMaxLine(AccordionDrawer.Y);
	}
	
	// don't draw self, just call this on parents after calling drawDescend
	// ascend drawing until the root is drawn or the node drew in the frame already
	public void drawAscend(int frameNum, float plane)
	{
		if (cell.getDrawnFrame() < frameNum)
		{
			drawInCell(cell.drawer.getColorsForCellGeom(this), plane);
			if (!isRoot())
				parent.drawAscend(frameNum, plane);
			cell.setDrawnFrame(frameNum);
		}
	}
	
	// don't draw self, just call this
	public void drawDescend(int frameNum, float plane, int min, int max)
	{
		drawInCell(cell.drawer.getColorsForCellGeom(this), plane);
		TreeNode currNode = this;
		// binary search
		int minChild = 0;
		int maxChild = currNode.children.size() - 1;
		int midChild = (minChild + maxChild) / 2;
		while (!currNode.isLeaf())
		{
			TreeNode midNode = currNode.getChild(midChild);
			if (midNode.key <= max && // mid node is either some tree above max or is an ancestor of max 
					midNode.rightmostLeaf.key >= min) // midnode has some leaf bigger than min 
			{
				midNode.drawInCell(cell.drawer.getColorsForCellGeom(midNode), plane);
				currNode = midNode;
				minChild = 0;
				maxChild = currNode.children.size() - 1;
			}
			else if (midNode.key <= max) // mid node too small, make it bigger
				minChild = midChild+1;
			else
				maxChild = midChild-1;
			midChild = (minChild + maxChild) / 2;
		}		
	}

	/* (non-Javadoc)
	 * @see AccordionDrawer.CellGeom#drawInCell(double)
	 */
	public void drawInCell(double plane, boolean doFlash) {
		// TODO Auto-generated method stub
		
	}

	/**
	 * Draws this TreeEdge inside the GridCell to which it is
	 * attached, and its label if appropriate.
	 *
	 * Find font size to use for drawing the label, by checking for
	 * occlusions. Don't draw it all if it's occluded even at the
	 * smallest size.
	 * 
	 * @author   Tamara Munzner, Li Zhang
	 * 
	 * @param    col The color to draw this node in
	 * @param    plane Z depth at which to draw label
	 *
	 * @see      AccordionDrawer.GridCell
	 */
	
	private Color setGLColor(AccordionTreeDrawer atd, Color col)
	{
		float rgbcol[] = new float[3];
		Color newrgb;
		if (atd.dimcolors || atd.dimbrite) {
			
			col.getRGBColorComponents(rgbcol);
			float hsbcol[] = Color.RGBtoHSB((int) (rgbcol[0] * 255),
					(int) (rgbcol[1] * 255), (int) (rgbcol[2] * 255),
					(new float[3]));
			float howdim;
			// H stays the same
			// if started out greyscale, don't crank up S
			if (hsbcol[1] > 0f && atd.dimcolors) {
				// depends on size of your subtended subtree screen area
				InteractionBox b = atd.makeBox(this); // box around subtree
				if (null == b)
					System.out.println("interaction box null");
				// howdim base: one pixel 0, fullscreen 1.0
				howdim = (float) (b.getMax(AccordionDrawer.Y) - b
						.getMin(AccordionDrawer.Y))
						/ (float) atd.getWinMax(1); //screen coords
				// small dynamic range:, 0%=.5, 20% and up =1.0
				howdim = .5f + 40 * howdim;
				howdim = (howdim > 1.0) ? 1.0f : howdim;
				hsbcol[1] = (float) howdim;
			}
	
			if (atd.dimbrite && col == atd.getObjectColor()) {
				// depends on depth in tree
				int treeheight = atd.tree.getHeight();
				// howdim ranges from 0 at root to 1.0 at leaf
				howdim = (height - 1.0f) / (treeheight - 1.0f);
				// compress the range
				hsbcol[2] = (float) (0.2 + howdim * .7);
			}
			newrgb = Color.getHSBColor(hsbcol[0], hsbcol[1], hsbcol[2]);
		} else {
			newrgb = col;
		}
		return newrgb;
	}

	public void setPositions(float[] start, float[] end, boolean horiz)//, int direction)
	{
		AccordionTreeDrawer atd = (AccordionTreeDrawer)cell.drawer;
		int X = AccordionDrawer.X, Y = AccordionDrawer.Y;
		GridCell c = getCell();
		double[] minStuck = {atd.getSplitLine(Y).getMinStuckValue(),
			atd.getSplitLine(Y).getMinStuckValue()};
		double[] rangeSize = {
			atd.getSplitLine(X).getMaxStuckValue() - minStuck[X],
			atd.getSplitLine(Y).getMaxStuckValue() - minStuck[Y]};
		if (horiz)
		{
			start[X] = (float)c.getMin(X); // X is over the range
			end[X] = (float)c.getMax(X);
			start[Y] = end[Y] = (float)getMidY();
		}
		else
		{
			start[X] = end[X] = (float)c.getMax(X);
			start[Y] = (float)getMinY();
			end[Y] = (float)getMaxY();
		}
	
	}

	/** 
	     * Create a LabelBox for the given fontheight
	     * @param    x horizontal base location in screen/pixel coordinates
	     * @param    y vertical base location in screen/pixel coordinates
	     * @author Tamara Munzner
	     * @see	 AccordionDrawer.LabelBox
	     */
	
	    protected LabelBox makeLabelBox(int fontheight, int x, int y, float[] start, float[] end) {
		    AccordionTreeDrawer d = (AccordionTreeDrawer)getCell().drawer;
		    int X = AccordionDrawer.X, Y = AccordionDrawer.Y;
			int bottomLeft[] = new int[2];
			int topRight[] = new int[2];
			int bottomLeftBackground[] = new int[2];
			int startPix[] = {d.w2s(start[X], X), d.w2s(start[Y], Y)};
			int endPix[] = {d.w2s(end[X], X), d.w2s(end[Y], Y)};
			String name = label;
			int namewidth = AccordionDrawer.stringWidth(name, fontheight); // expensive
			int oldnamewidth = 0;
			int labelwidth = namewidth+2;
			int numChars;
			int[] winSize = {d.getWinsize(X), d.getWinsize(Y)};
	
			int moveover = (isLeaf()) ? labelwidth + d.labeloffset[X] : -d.labeloffset[X];
	//			if(d.drawDirection != AccordionDrawer.RIGHT)
				{
					topRight[X] = (d.labelposright) ?
					endPix[X]+moveover :
					startPix[X]-d.labeloffset[X];
				}
			// either put label right and above junction
			// or put it left, vertically centered, and right-justified
			if (!d.labelposright)
			{
	//			if(d.leftRight == 0 || d.leftRight == 1)
				topRight[Y] = startPix[Y]-d.labeloffset[Y]-fontheight;
	//		   if(d.leftRight == 2)
	//		   labelTopRight[1] = d.w2s(1-posStart[0],1)-d.labeloffset[1]-fontheight;
	//		   if(d.leftRight == 3)
	//			labelTopRight[1] = d.w2s(posStart[0],1)-d.labeloffset[1]-fontheight;
			}
			else 				
				topRight[Y] = (isLeaf()) ? 
						endPix[Y] - (int)(fontheight/2.0) : 
						endPix[Y]-d.labeloffset[Y] - fontheight;       		
			
			// we pad background label box by one pixel on each side so
			// that text is never right up against the edge. height
			// already has enough padding
			int labelheight = fontheight+2;	
			bottomLeftBackground[X] = topRight[X] - labelwidth;
			bottomLeftBackground[Y] = topRight[Y] + labelheight;
			
		
			if (x>=0 && y>=0 && d.labelpopup) {
				bottomLeftBackground[X] = x-(int)(labelwidth/2.0);
				topRight[X] = bottomLeftBackground[X]+labelwidth;
				bottomLeftBackground[Y] = y-5;
				topRight[Y] = bottomLeftBackground[Y]-labelheight;
			} 
	
			// if we would start offscreen, nudge over until it's all
			// onscreen. when the origin of the text string is offscreen,
			// opengl won't draw anything (sigh). we could truncate the
			// label text, but it seems more useful to just move the whole
			// thing over. while we're at it, make sure labels are always
			// totally onscreen.
	
			int labelToEdge = winSize[X] - endPix[X];
			if ((bottomLeftBackground[X]) < 0) {
			   bottomLeftBackground[X] = 0;
			   topRight[X] = bottomLeftBackground[X] + labelwidth;
			} else if (topRight[X] > winSize[X]) {	
				// do nothing - in leaf case, we want to truncate not nudge
				// as per hillis feedback.
			} 
	//		System.out.println("top right is less than posend  ############"+labelTopRight[0]);
	//		 System.out.println("pos end  ############" + d.w2s(1-posEnd[0], 0));
	
			if (bottomLeftBackground[Y] > winSize[Y]) {
				bottomLeftBackground[Y] = winSize[Y];
				topRight[Y] = bottomLeftBackground[Y]-labelheight;
			} else if (topRight[Y] < 0) {
				topRight[Y] = 0;
				bottomLeftBackground[Y] = labelheight;
			} 
	
			// labelbuffer is used for occlusion computations to control
			// label density, but we don't want to draw the bg box that
			// big because it would look silly.
	
			bottomLeft[Y] = bottomLeftBackground[Y] + d.labelbuffer[Y];
			bottomLeft[X] = bottomLeftBackground[X] - d.labelbuffer[X];
			return new LabelBox( bottomLeft, topRight, bottomLeftBackground, d.getFrameNum(), name);
	  }

	/** 
	 * Occlusion check of LabelBox against array of all drawn labels. 
	 * {@link #drawnArray AccordionDrawer.drawnArray} is sorted in y -
	 * the skinny direction, to avoid work. We do a binary search to
	 * check for occlusions, so it's log in the number of pixels (with
	 * a very small constant).
	 * 
	 * @author  Tamara Munzner
	 * @see AccordionDrawer.LabelBox
	 */
	private boolean intersectLabelBox(LabelBox lb) {
	AccordionTreeDrawer d = (AccordionTreeDrawer)getCell().drawer;
	
	// the list is sorted by y value. use binary search to see
	// if it will fit instead of checking against every item. 
	// find an item that has y overlap. check down from there
	// until no overlap, then check up until no overlap. stop
	// as soon as there's overlap, of course.
	boolean intersects = false;
	ArrayList drawnLabels = d.getDrawnLabels();
	int listlength = drawnLabels.size();
	//int high = listlength-1;
	int high=(listlength==1)?listlength: listlength-1;
	int low = 0;
	int current = -1;
	int prevcurrent = -2;
	int foundIndex = -1;
	LabelBox found;
	//while ((low < high) && (current != prevcurrent) && (!intersects)) {
	while ((low < high) && (!intersects) && (current<=high) && (foundIndex < 0))
	{
	    prevcurrent = current;
	    current = (int)((high+low)/2.0);
	    if (current == prevcurrent && low+1 == high) {
			low=high;
			current = high;
	    }
		if(current == listlength) current = current-1;
	    LabelBox prevlab = (LabelBox) drawnLabels.get(current);
	    //System.out.println("lb bl " + lb.bl[1] + " p tr " + prevlab.tr[1] + " lb tr " + lb.tr[1] + " p bl " + prevlab.bl[1]);
	    if (lb.compareBtoT(prevlab, low, high)) high = current;
	    else if (prevlab.compareBtoT(lb, low, high)) low = current;
	    else { // check intersections
		found = prevlab;
		foundIndex = current;
		//System.out.println("found: " + foundIndex);
		// we're doing a bit more work than necessary by
		// checking against prevlab from last iteration,
		// but it's the easiest way to ensure corner cases
		// work
		while (!intersects && prevlab.bottomBiggerThanTop(lb) && current >= 0) { 
		    // check down from found
		    // i.e. index in sorted list gets smaller, so
		    // y values get smaller also
		    prevlab = (LabelBox) drawnLabels.get(current);
		    intersects = lb.intersectBoxes(prevlab);
		    current--;
		    //System.out.println("cur-- now " + current);
		}
		//current = foundIndex+1;
		current = foundIndex;
		while (!intersects && lb.bottomBiggerThanTop(prevlab) && current <= listlength-1) { 
		    // check up from found
		    prevlab = (LabelBox) drawnLabels.get(current);
		    intersects = lb.intersectBoxes(prevlab);
		    current++;
		    //System.out.println("cur++ now " + current);
		}
	    }
	    //System.out.println("low " + low + " high " + high + " cur " + current + " prevcur " + prevcurrent + " intersects " + intersects);
	}
	return intersects;
	}

	/** Draw a LabelBox. 
	     * @author	 Tamara Munzner, Li Zhang
	     * @param	 lb LabelBox to use
	     * @param    fontheight Size of font to use, in points/pixels.
	     * @param	 drawBig whether to draw maximum size ignoring occlusions
	     */
	    private void drawLabelBox(LabelBox lb, int fontheight, boolean drawBig) 
	    {
	    	int X = AccordionDrawer.X, Y = AccordionDrawer.Y;
			AccordionTreeDrawer d = (AccordionTreeDrawer)getCell().drawer;
			GL gl = (GL) d.getGraphicsStuff();
			ArrayList drawnLabels = d.getDrawnLabels();
			if (!drawBig)
			{
		    	drawnLabels.add(lb);
		    	Collections.sort(drawnLabels);
			}
			
			int[] pos = { lb.pos(0), lb.pos(1) };
			int[] topRightPos = { lb.topRightPos(0), lb.topRightPos(1) };
	
			// highlighted node drawing == drawBig
	
			double labelplane = d.getLabelplane();
			if (d.labeldrawback || drawBig) {
				float thecol[] = new float[3];
				if (drawBig)
					d.getLabelBackHiColor().getRGBColorComponents(thecol);
				else
					d.getLabelBackColor().getRGBColorComponents(thecol);
				if (!drawBig && d.labeltransp) {
					gl.glColor4f(thecol[0], thecol[1], thecol[2], .5f);
				} else {
					gl.glColor3f(thecol[0], thecol[1], thecol[2]);
				}
	//			if (d.drawDirection != AccordionDrawer.RIGHT) {
					gl.glBegin(GL.GL_POLYGON);
					gl.glVertex3d(d.s2w(pos[0], 0), d.s2w(pos[1], 1), labelplane);
					gl.glVertex3d(d.s2w(topRightPos[0], 0), d.s2w(pos[1], 1),
							labelplane);
					gl.glVertex3d(d.s2w(topRightPos[0], 0),
							d.s2w(topRightPos[1], 1), labelplane);
					gl.glVertex3d(d.s2w(pos[0], 0), d.s2w(topRightPos[1], 1),
							labelplane);
					gl.glEnd();
	
			}
			int descent = AccordionDrawer.getDescent(fontheight);
			String name = lb.getName();
			double b[] = { d.s2w(pos[X] + 1, X), d.s2w(pos[Y] - descent - 1, Y)};
			if (drawBig) // no shadow on highlight
			{
				d.setColorGL(d.getLabelHiColor());
				d.drawText(b[X], b[Y], name, fontheight, d.getLabelColor(), labelplane, null);
			}
			else if (!d.labeldrawback) // shadow, not big, or background box
			{
				d.drawText(b[X], b[Y], name, fontheight, d.getLabelColor(), labelplane, d.getLabelBackColor());
			}
			else // not big, background box so no shadow needed
				d.drawText(b[X], b[Y], name, fontheight, d.getLabelColor(), labelplane, null);
		}

	
}
