
/*
   Copyright (c) 2002 Compaq Computer Corporation
   
   SOFTWARE RELEASE
   
   Permission is hereby granted, free of charge, to any person obtaining
   a copy of this software and associated documentation files (the
   "Software"), to deal in the Software without restriction, including
   without limitation the rights to use, copy, modify, merge, publish,
   distribute, sublicense, and/or sell copies of the Software, and to
   permit persons to whom the Software is furnished to do so, subject to
   the following conditions:
   
   - Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
   
   - Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
   
   - Neither the names of Compaq Research, Compaq Computer Corporation
     nor the names of its contributors may be used to endorse or promote
     products derived from this Software without specific prior written
     permission.
   
   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
   IN NO EVENT SHALL COMPAQ COMPUTER CORPORATION BE LIABLE FOR ANY CLAIM,
   DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
   OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
   THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

package TreeJuxtaposer;

import AccordionDrawer.AccordionDrawer;
import AccordionTreeDrawer.*;
//import AccordionTreeDrawer.AccordionTreeDrawer;
//import AccordionTreeDrawer.Tree;
//import AccordionTreeDrawer.TreeNode;

import java.util.*;


/**
 * Tree2Tree does the precomputation for each pair of trees. The two
 * precomputation tasks are computing the best corresponding node for
 * each node and building the range query data structures.
 *
 * @author  Tamara Munzner, Serdar Tasiran, Li Zhang, Yunhong Zhou
 * @version 
 * @see     AccordionDrawer.Tree
 * @see     TreeJuxtaposer.RangeTree
 */

class Tree2Tree {

	// the trees being compared
    Tree treeA, treeB;

	// for X2Y: an entry for each node in X if the subtree under X is a forest in Y
	Hashtable A2B, B2A;

    /**
     * The hashmaps that store the best matchings. A2B (B2A) stores
     * the best match of treeB (treeA) node to treeA (treeB).
     */

    // the vector that stores the best corresponding nodes for
    // different levels. each element of the vector is a hashmap that
    // is indexed by nodes
    Vector bestA2B, bestB2A;

    // to deal with edge weights and leaves
    float epsilon = 0.1f;

    //    HashMap bestA2B, bestB2A;
//    RangeTree rangeAB, rangeBA;
    
    // do all the necessary preprocessing
    Tree2Tree(Tree t1, Tree t2, int eL) {
	treeA = t1;
	treeB = t2;
	long start, current;
	bestA2B = new Vector(eL);
	bestB2A = new Vector(eL);

	start = System.currentTimeMillis();
	// initialization the data structures for answering the
	// queries of getBestCorrNode and isRangeInRange.
	computeBestMatch(treeA, treeB, bestA2B, eL);
	current = System.currentTimeMillis();
	if (AccordionDrawer.debugOutput)
	    System.out.println("best match 1 total preprocessed: "+(current-start)/1000.0f+" sec");
	start = current;
	computeBestMatch(treeB, treeA, bestB2A, eL);
	current = System.currentTimeMillis();
	if (AccordionDrawer.debugOutput)
	    System.out.println("best match 2 total preprocessed: "+(current-start)/1000.0f+" sec");
	start = current;
	
	// initialize the data structure for rectangular range searching
//	rangeAB = new RangeTree(treeA, treeB, this);
//	rangeBA = new RangeTree(treeB, treeA, this);
//	current = System.currentTimeMillis();
//	System.out.println("range tree preprocessed: "+(current-start)/1000.0f+" sec");

    }

	/**
	 * clean method, called when a tree is deleted
	 * @see TreePairs.removeTree
	 *
	 */
		public void close(){
		   if (A2B != null) A2B.clear();
		   if (B2A != null) B2A.clear();	
		   if (bestA2B != null) bestA2B.clear();
		   if (bestB2A != null) bestB2A.clear();
//		   this = null;
		}

	private void addNodeToForest(TreeNode node, ArrayList array, Hashtable hash)
	{
		Integer nodeInteger = new Integer(node.key);
		if (hash.get(nodeInteger) == null)
		{
			array.add(nodeInteger);
			hash.put(nodeInteger, node);
		}
	}
	private void removeNodeFromForest(TreeNode node, ArrayList array, Hashtable hash)
	{
		Integer nodeInteger = new Integer(node.key);
		if (hash.get(nodeInteger) == null)
		{
			array.remove(nodeInteger);
			hash.remove(nodeInteger);
		}
	}

	/*
	 * return an ArrayList of a reduced number of elements from ArrayList
	 * based on the number of leaves; bigger subtrees are kept, ties broken
	 * by taking any subtree with maximum number of leaves not already in forest
	 */
	private ArrayList reduceNodeListToCutoff(ArrayList node, int cutoff)
	{
		//System.out.println("reducing nodelist to cutoff");
		int subtreeSize[] = new int[node.size()];
		for (int i = 0; i < node.size(); i++)
		{
			subtreeSize[i] = ((TreeNode)node.get(i)).numberLeaves;
		}
		Arrays.sort(subtreeSize); // sort array in ascending order
		int subtreeSizeCutoff = subtreeSize[node.size() - cutoff] + 1; // find (cutoff + 1) subtree size
		ArrayList returnList = new ArrayList();
		for (int i = 0; i < node.size(); i++)
		{
			TreeNode currentNode = (TreeNode)node.get(i);
			if (currentNode.numberLeaves >= subtreeSizeCutoff)
				returnList.add(currentNode);
		}
		subtreeSizeCutoff--;
		for (int i = 0; i < node.size() && returnList.size() < cutoff; i++)
		{
			TreeNode currentNode = (TreeNode)node.get(i);
			if (currentNode.numberLeaves == subtreeSizeCutoff)
				returnList.add(currentNode);
		}
		return returnList;
	}
	
	private void computeForest(Hashtable X2Y, Tree treeX, Tree treeY,
		AccordionTreeDrawer atdY, int cutoff)
	{
		//posorder = children first, then parents
		for (TreeNode nX = treeX.posorderStartNode; nX != null; nX = nX.posorderNext)
		{
			ArrayList currListX = null;
			TreeNode nY = getBestCorrNode(treeX, nX, treeY, 0);
			if (nY == null) // no BCN for nX, give up
				continue;
			Hashtable nXHash = null;

			if (!nX.isLeaf()) // add children of nY to currListY
			//  (iff child is BCN of a descendent of nX)
			{
				currListX = new ArrayList();
				nXHash = new Hashtable();
				addNodeToForest(nY, currListX, nXHash);
				// add nY reference (sort later)
				for (int i = 0; i < nX.numberChildren(); i++) {
					TreeNode nXChild = nX.getChild(i);
					TreeNode nYChild = getBestCorrNode(treeX, nXChild, treeY, 0);
					ArrayList nXChildList = (ArrayList)X2Y.get(nXChild);
					if (nXChildList != null) {
						for (int j = 0; j < nXChildList.size(); j++) // go through forest of nXChild
						{
							TreeNode nYj = (TreeNode)nXChildList.get(j);
							if (nY.getMin() > nYj.getMin() || nY.getMax() < nYj.getMax())
							// nY is not an ancestor of nYj; nYj could be an ancestor of nY
								addNodeToForest(nYj, currListX, nXHash);
							if (nYj.getMin() <= nY.getMin() && nYj.getMax() >= nY.getMax())
							// nYj is an ancestor of nY, remove nY
								removeNodeFromForest(nY, currListX, nXHash);
						}
					}
					// add child
					if (nYChild != null && !nY.isAncestorOf(nYChild))
						addNodeToForest(nYChild, currListX, nXHash);
						
					// reduce (to cutoff) large forests under children of current node
					if (nXChildList != null && nXChildList.size() > cutoff)
					{
						ArrayList newNXChildList = reduceNodeListToCutoff(nXChildList, cutoff);
						X2Y.remove(nXChild);
						X2Y.put(nXChild, newNXChildList);
					}
				}
			}
			ArrayList finalArrayX = new ArrayList(); // final forest calculation
			
			// sort nBChildList by key
			if (currListX != null) {
				Object tempObjects[] = currListX.toArray();
				Integer tempArray[] = new Integer[tempObjects.length];
				for (int i = 0; i < tempObjects.length; i++)
				  tempArray[i] = (Integer)tempObjects[i];
				Arrays.sort(tempArray);
				for (int i = 0; i < tempArray.length; i++)
				{
					TreeNode node = (TreeNode)nXHash.get(tempArray[i]);
					if (node != null)
					{
						finalArrayX.add(node);
						nXHash.remove(tempArray[i]);
					}
				}
			}
			
			// add finalArrayX to X2Y referenced by nX if finalArrayX has more than one element
			if (finalArrayX != null && finalArrayX.size() > 1)
				X2Y.put(nX, finalArrayX);
		} // end of current nX
	}


	/*
	 * preprocessing: calculate and store forests that correspond to subtrees
	 * A2B/B2A: hash tables that map nodes (roots of subtrees) from A/B to forests of nodes (subtrees) in B/A
	 * all work is done in computeForest
	 */
	public void subtree2Forest(AccordionTreeDrawer atdA, AccordionTreeDrawer atdB, int eL)
	{
		final int cutoff = 100; // allow at most "cutoff" items in a node's forest array
		float start = System.currentTimeMillis();
		A2B = new Hashtable();
		computeForest(A2B, treeA, treeB, atdB, cutoff);
		B2A = new Hashtable();
		computeForest(B2A, treeB, treeA, atdA, cutoff);
		float time = (System.currentTimeMillis() - start) / 1000;
		if (AccordionDrawer.debugOutput)
		    System.out.println("Time to preprocess forest pair: " + time);
	}

    /**
     * Computes the node in Tree "other" whose set of descendant
     * leaves best matches that of TreeNode n in Tree "source"
     * 
     * The best match is the node n' maximizing the following score
     * | S(n) Intersection S(n') | / | S(n) Union S(n') | 
     * 
     * where S(n) is the set of leaves that are descendants of node n.
     * 
     * @author   Tamara Munzner, Serdar Tasiran, Li Zhang, Yunhong Zhou
     * 
     * @see      AccordionDrawer.Tree
     * @see      AccordionDrawer.TreeNode
     * @see      TreeJuxtaposer.NodeScorePair
     */

    TreeNode getBestCorrNode(Tree source, TreeNode n, Tree other, int el) {
        HashMap h = null;
	if((source == treeA)&&(other == treeB))
	{
	    h = (HashMap)bestA2B.elementAt(el);
	}
	else if((source == treeB)&&(other == treeA))
	{
	    h = (HashMap)bestB2A.elementAt(el);
	}
	if (h == null) return null;
	
    NodeScorePair p = ((NodeScorePair)(h.get(n)));
    if (p != null)
        return p.node;
    return null;
    }

    float getBestCorrNodeScore(Tree source, TreeNode n, Tree other, int el) {

	if((source == treeA)&&(other == treeB))
	    return ((NodeScorePair)(((HashMap)(bestA2B.elementAt(el))).get(n))).score;
	else if((source == treeB)&&(other == treeA))
	    return ((NodeScorePair)(((HashMap)(bestB2A.elementAt(el))).get(n))).score;
	else return -1.0f;
    }

	// return an ArrayList of RangeInTree objects that correspond to n from source
	ArrayList getCorrRange(Tree source, TreeNode n, Tree other, int el)
	{
		int min = n.getMin();
		if ((source == treeA)&&(other == treeB))
			return (ArrayList)A2B.get(n);
		else if ((source == treeB) && (other == treeA))
			return (ArrayList)B2A.get(n);
		return null;
	}


//    /** 
//     * Given a node range in one tree, say whether there's an overlap
//     * with the node range in the other tree. 
//     * Returns number of overlapping nodes, possibly 0
//     * 
//     * @author   Tamara Munzner, Serdar Tasiran, Li Zhang, Yunhong Zhou
//     * 
//     * @see      AccordionDrawer.Tree
//     * @see      AccordionDrawer.TreeNode
//     */
//
//    int isRangeInRange(Tree source, int AMin, int AMax, 
//		       Tree other, int BMin, int BMax) throws Exception{
//	if((source == treeA)&&(other == treeB)) 
//	    return rangeAB.query2D(AMin, AMax, BMin, BMax);
//	else if((source == treeB)&&(other == treeA)) 
//	    return rangeBA.query2D(AMin, AMax, BMin, BMax);
//	else
//	    return -1;
//    }

    class NodeScorePair {

	TreeNode node = null;
	float score = 0.0f;
	NodeScorePair(TreeNode n, float s) { node = n; score = s; }
    };

    /** 
     * Attachment to a node that is needed as temporary data structure
     * when computing best corresponding nodes
     **/
    class TmpD {
	int tmpScore = 0;
	float uSum = 0;
	float lSum = 0;
	TreeNode tmpParent = null;
	TreeNode tmpPosorderNext = null;
	TmpD() {;}
    };

    /**
     *
     * For each node om Tree t1, computes the best matching node in
     * Tree t2 and stores it in HashMap h12
     * 
     * @see      AccordionDrawer.Tree
     * @see      TreeJuxtaposer.computeBestMatch
     * @see      TreeJuxtaposer.NodeScorePair
     */

    
    void computeBestMatch(Tree t1, Tree t2, Vector v12, int eL) {
	
	TmpD[] tmpData = new TmpD[t2.nodes.size()];

	// allocate the temporary data needed. we will reuse
	// the temporary data for each node. hopefully, this saves
	// us some time.
	for(int i=0; i<t2.nodes.size(); i++) {
	    tmpData[i] = new TmpD();
	}
	HashMap h12 = new HashMap();

	long start, current;
	// special case for level 0 since we know how to deal with 
	// it
	start = System.currentTimeMillis();
	for(int i=0; i<t1.nodes.size(); i++) {
	    TreeNode n = (TreeNode) t1.nodes.get(i);
	    NodeScorePair p = computeBestMatch(n, t1, t2, tmpData);
	    h12.put(n, p);
	}
	current = System.currentTimeMillis();
	if (AccordionDrawer.debugOutput)
	    System.out.println("best match level 0 preprocessed: "+(current-start)/1000.0f+" sec");
	start = current;

	v12.add(h12);
	for(int el=1; el<eL; el++) {
	    h12 = new HashMap();
	    for(int i=0; i<t1.nodes.size(); i++) {
		TreeNode n = (TreeNode) t1.nodes.get(i);
		NodeScorePair p = computeBestMatch(n, t1, t2, el*1.0f/(eL-1), tmpData);
		h12.put(n, p);
		// System.out.println(n.leftmostLeaf.getName()+","+n.rightmostLeaf.getName()+":"+p.node.leftmostLeaf.getName()+","+p.node.rightmostLeaf.getName()+" "+p.score);
	    }
	    current = System.currentTimeMillis();
	    if (AccordionDrawer.debugOutput)
	        System.out.println("best match level "+el+" preprocessed: "+(current-start)/1000.0f+" sec");
	    start = current;
	    v12.add(h12);
	}
    }

    /**
     * Computes the dissimilarity distance between treeA and treeB
     * 
     * The distance is computed by summing up getBestCorrNode scores
     * between each tree node in treeA and the best matching node in
     * treeB. 
     *
     * <Pre>
     * mode 0: alpha is the cut off
     * 1: (1-s)^\alpha
     * 2: 1-s^{1/\alpha}
     * </Pre>
     *
     */ 

    float computeDistance(float alpha, int mode) {
	float dif1 = 0.0f;
	float s = 0.0f;
	for(int i=0; i<treeA.nodes.size(); i++) {

	    try {
		s = getBestCorrNodeScore(treeA, (TreeNode) treeA.nodes.get(i), treeB, 0);} 
	    catch (Exception e) {
		
	    }
	    if(alpha==0) {
		if(s<1.0f) dif1+=1.0f;
	    } else if (alpha==1) {
		dif1 += (1-s);
	    } else {	   
		if(mode == 0) {
		    s = 1.0f - s; // convert similarity to distance
		    if(s>alpha) dif1+=1.0;
		    else dif1 += s;
		} else if(mode == 1) {
		    s = 1.0f - s; // convert similarity to distance
		    dif1 += Math.pow(s,alpha);
		} else if (mode == 2) {
		    dif1 += (1-Math.pow(s, 1.0/alpha));
		}
	    }
	}

	float dif2 = 0.0f;
	for(int i=0; i<treeB.nodes.size(); i++) {
	    try {
		s = getBestCorrNodeScore(treeB, (TreeNode) treeB.nodes.get(i), treeA, 0);}
	    catch (Exception e) {
		
	    }
	    if(alpha==0) {
		if(s<1.0f) dif2+=1.0f;
	    } else if (alpha==1) {
		dif2 += (1-s);
	    } else {	   
		if(mode == 0) {
		    s = 1.0f - s; // convert similarity to distance
		    if(s>alpha) dif2+=1.0;
		    else dif2 += s;
		} else if(mode == 1) {
		    s = 1.0f - s; // convert similarity to distance
		    dif2 += Math.pow(s,alpha);
		} else if (mode == 2) {
		    dif2 += (1-Math.pow(s, 1.0/alpha));
		}
	    }
	}
	return (dif1>dif2?dif1:dif2);
    }
    
    
    /**
     * compute the best corresponding node for each node
     *
     * node B is the best corresponding node of node A if it maximizes
     *
             | L(A) U L(B)|
            ----------------
             | L(A) n L(B)|
	  
	  where L(A),L(B) represent the set of leaves underneath the
	  node A and node B respectively.
	  
	  For the description of the algorithm, see 
	  Li Zhang. On Matching Nodes in Two Trees.
    **/
    NodeScorePair computeBestMatch(TreeNode sourceNode, Tree sourceTree,
    	Tree targetTree, TmpD[] tmpData) {

	LinkedList mleaves = new LinkedList();
	Object[] la;
//	TreeNode n, nn;

	for(int currLeafIndex = sourceNode.leftmostLeaf.getLindex();
		currLeafIndex <= sourceNode.rightmostLeaf.getLindex();
		currLeafIndex++) {
			TreeNode n = sourceTree.getLeaf(currLeafIndex);
	    	TreeNode nn = targetTree.getNodeByName(n.getName());
	    if(nn!=null) {
		if(nn.isLeaf()) {
		    tmpData[nn.key].tmpScore = 1;
		    mleaves.add(nn);
		}
	    }
	}	
	if(mleaves.size()==0) {
	    return new NodeScorePair(null, 0.0f);
	} else if(mleaves.size()==1) {
	    return new NodeScorePair((TreeNode)(mleaves.get(0)), 1.0f/sourceNode.numberLeaves);
	}

	la = mleaves.toArray();

	Arrays.sort(la, new LeafComparator());
	
	int len = la.length;

	// build the simplified spanning tree by repeated bay traversal.
	// the following precedure can be probably best understood by 

	TreeNode tmpPosorderStart = (TreeNode)(la[0]);
	TreeNode lastNode = (TreeNode)(la[len-1]);
	TreeNode tmpRoot = tmpPosorderStart.leafLca(lastNode);
	
	tmpData[tmpPosorderStart.key].tmpParent = tmpRoot;
	tmpData[lastNode.key].tmpParent = tmpRoot;	
	tmpData[tmpPosorderStart.key].tmpPosorderNext = lastNode;
	tmpData[lastNode.key].tmpPosorderNext = tmpRoot;

	Stack bay = new Stack();
	TreeNode prev, pprev,a;

	bay.push(tmpRoot);
	bay.push(tmpPosorderStart);

	for(int i=1; i<len-1; i++) {

	    TreeNode nn = (TreeNode) (la[i]);
	    if(nn.getLindex() == ((TreeNode)la[i-1]).getLindex())
		continue; // no duplication, please
	    a = nn.leafLca((TreeNode)(la[i-1]));
	    prev = null;
	    pprev = null;

	    while(!bay.empty()) {
		pprev = prev;
		prev = (TreeNode) (bay.peek());
		if(prev.isAncestorOf(a)) {
		    if(prev==a) bay.pop();
		    break;
		}
		bay.pop();
	    }

	    if(bay.empty()) {
		a = nn.leafLca(lastNode);
		if(a==prev) {
		    // not a binary tree
		    tmpData[nn.key].tmpParent = prev;
		    tmpData[pprev.key].tmpPosorderNext = nn;

		    // we need to deal with non-binary tree and when
		    // the node inserted is a child of the root
		    if(prev==tmpRoot)
			tmpData[nn.key].tmpPosorderNext = lastNode;
		    else
			tmpData[nn.key].tmpPosorderNext = a;
		} else {
		    
		    // the node is inserted to the right branch
		    // and creat the new tmpRoot
		    tmpData[a.key].tmpParent = tmpData[lastNode.key].tmpParent;
		    tmpData[lastNode.key].tmpParent = a;
		    tmpData[nn.key].tmpParent = a;
		    tmpData[a.key].tmpPosorderNext = 
			tmpData[lastNode.key].tmpPosorderNext;
		    tmpData[pprev.key].tmpPosorderNext = nn;
		    tmpData[nn.key].tmpPosorderNext = lastNode;
		    tmpData[lastNode.key].tmpPosorderNext = a;
		    tmpRoot = a;
		}
	    } else {
		if(a==prev) {
		    tmpData[nn.key].tmpParent = prev;
		    tmpData[pprev.key].tmpPosorderNext = nn;
		    tmpData[nn.key].tmpPosorderNext = a;
		} else {
		    tmpData[a.key].tmpParent = tmpData[pprev.key].tmpParent;
		    tmpData[pprev.key].tmpParent = a;
		    tmpData[nn.key].tmpParent = a;
		    
		    tmpData[a.key].tmpPosorderNext = tmpData[pprev.key].tmpPosorderNext;
		    tmpData[pprev.key].tmpPosorderNext = nn;
		    tmpData[nn.key].tmpPosorderNext = a;
		}
	    }
	    bay.push(a);
	    bay.push(nn);
	}

	// walk up in the tree, compute the score of each node in the
	// spanning tree, and find the maximum score
	TreeNode match = null;
	float matchScore = 0.0f;
	float currentMatchScore;
	int sizeofUnion;

	for(TreeNode n = tmpPosorderStart;
		n!=null;
		n=tmpData[n.key].tmpPosorderNext) {

	    sizeofUnion = sourceNode.numberLeaves+n.numberLeaves-tmpData[n.key].tmpScore;
	    currentMatchScore = (float)(tmpData[n.key].tmpScore*1.0/sizeofUnion);
	    if(matchScore<currentMatchScore) {
		match = n; 
		matchScore = currentMatchScore;
	    }
	    if(matchScore == 1.0f) break;
	    TreeNode np = tmpData[n.key].tmpParent;
	    if(np != null) {
		tmpData[np.key].tmpScore += tmpData[n.key].tmpScore;
	    }
	}
	
	// cleanup the temporary data
	TreeNode n = tmpPosorderStart;
	while(n!=null) {
	    tmpData[n.key].tmpScore = 0;
	    TreeNode nn = tmpData[n.key].tmpPosorderNext;
	    tmpData[n.key].tmpPosorderNext = null;
	    tmpData[n.key].tmpParent = null;
	    n = nn;
	}
	return new NodeScorePair(match, matchScore);
    }


    // deal with edge weights

    NodeScorePair computeBestMatch(TreeNode sourceNode, Tree sourceTree,
    	Tree targetTree, float alpha, TmpD[] tmpData) {
	
	// compute the path length to each leaf node
	HashMap h = new HashMap();
	float tSum = 0;
	
	for (int currLeafIndex = sourceNode.leftmostLeaf.getLindex();
		currLeafIndex <= sourceNode.rightmostLeaf.getLindex();
		currLeafIndex++)
	{
		TreeNode currSourceLeaf = sourceTree.getLeaf(currLeafIndex);
	    TreeNode p = currSourceLeaf;
	    float pathLen = epsilon;
	    while(p!=sourceNode)
	    {
			pathLen += p.weight;
			p = p.parent;
	    }
	    float pathLenA = (float)Math.pow(pathLen, alpha);
	    // float pathLenA = Math.exp(alpha*Math.log(pathLen));
	    h.put(currSourceLeaf.getName(), new Double(pathLenA));
	    tSum += pathLenA;
	}

	// prepare tmpData
	for(int i=0; i<tmpData.length; i++) {
	    TmpD tmpd = tmpData[i];
	    tmpd.uSum = 0;
	    // assume that the leaves are not presented in each node
	    // and then correct the assumption later.
	    tmpd.lSum = tSum;
	}

	// for each leaf node of t, accumulate the score bottom up
	// along the path to the root.
	for (int currTargetLeafIndex = 0;
		currTargetLeafIndex < targetTree.getLeafCount();
		currTargetLeafIndex++)
		{
		TreeNode currTargetLeaf = targetTree.getLeaf(currTargetLeafIndex);
	    float pathLen = epsilon;
	    Double tpl = (Double)(h.get(currTargetLeaf.getName()));
	    TreeNode p = currTargetLeaf;
	    if(tpl == null) {
		// the leaf is not in the subtree rooted at an
		while(p!=null) {
		    // tmpData[p.key].uSum += 0;
		    tmpData[p.key].lSum += Math.pow(pathLen, alpha);
		    // tmpData[p.key].lSum += Math.exp(alpha*Math.log(pathLen));
		    pathLen += p.weight;
		    p = p.parent;
		}
	    } else {
		float plen = tpl.floatValue();
		while(p!=null) {
		    TmpD tmpd = tmpData[p.key];
		    tmpd.lSum -= plen;
		    float plenp = (float)Math.pow(pathLen, alpha);
		    // float plenp = Math.exp(alpha*Math.log(pathLen));
		    if(plenp < plen) {
			tmpd.uSum += plenp;
			tmpd.lSum += plen;
		    } else {
			tmpd.uSum += plen;
			tmpd.lSum += plenp;
		    }
		    pathLen += p.weight;
		    p = p.parent;
		}

	    }

	}

	// traverse the tree and compute the best match and clean up
	// tmpData
	TreeNode match = null;
	float matchScore = 0.0f;
	float currentMatchScore;

	for(int i=0; i<targetTree.nodes.size(); i++) {
	    TreeNode n = (TreeNode)(targetTree.nodes.get(i));
	    TmpD tmpd = tmpData[n.key];
	    currentMatchScore = tmpd.uSum/tmpd.lSum;
	    if(matchScore<currentMatchScore) {
		match = n; 
		matchScore = currentMatchScore;
	    }
	    if(matchScore == 1.0f) break;
	}
	return new NodeScorePair(match, matchScore);
    }
};

